# Databricks notebook source
# MAGIC %pip install dbtunnel[asgiproxy,chainlit] databricks-genai-inference langchain-community tiktoken langchain mlflow-skinny chromadb rank_bm25 simsimd

# COMMAND ----------

dbutils.library.restartPython()

# COMMAND ----------

import os

current_directory = os.getcwd()
script_path = current_directory + "/chainlit_foundation_model.py"

# COMMAND ----------

from dbtunnel import dbtunnel
dbtunnel.chainlit(script_path).inject_auth().run()

# COMMAND ----------





import abc
import json
import os
from dataclasses import dataclass
from typing import Dict, Any, Literal, Optional
from urllib.parse import urlparse

from dbtunnel.utils import pkill


@dataclass
class ProxySettings:
    proxy_url: str
    port: str
    url_base_path: str
    url_base_path_no_port: Optional[str] = None


def get_cloud(context: Dict[str, Any]) -> str:
    # TODO: support gcp
    if context["extraContext"]["api_url"].endswith("azuredatabricks.net"):
        return "azure"
    return "aws"


def get_cloud_proxy_settings(cloud: str, org_id: str, cluster_id: str, port: int) -> ProxySettings:
    cloud_norm = cloud.lower()
    if cloud_norm not in ["aws", "azure"]:
        raise Exception("only supported in aws or azure")
    prefix_url_settings = {
        "aws": "https://dbc-dp-",
        "azure": "https://adb-dp-",
    }
    suffix_url_settings = {
        "aws": "cloud.databricks.com",
        "azure": "azuredatabricks.net",
    }
    # org_id = self._context["tags"]["orgId"]
    org_shard = ""
    # org_shard doesnt need a suffix of "." for dnsname its handled in building the url
    # only azure right now does dns sharding
    # gcp will need this
    if cloud_norm == "azure":
        org_shard_id = int(org_id) % 20
        org_shard = f".{org_shard_id}"
    # cluster_id = self._context["tags"]["clusterId"]
    url_base_path_no_port = f"/driver-proxy/o/{org_id}/{cluster_id}"
    url_base_path = f"{url_base_path_no_port}/{port}/"
    return ProxySettings(
        proxy_url=f"{prefix_url_settings[cloud_norm]}{org_id}{org_shard}.{suffix_url_settings[cloud_norm]}{url_base_path}",
        port=str(port),
        url_base_path=url_base_path,
        url_base_path_no_port=url_base_path_no_port
    )


Flavor = Literal[
    "gradio", "fastapi", "nicegui", "streamlit", "stable-diffusion-ui", "bokeh", "flask", "dash", "solara",
    "code-server", "chainlit"]


# from langchain: https://github.com/langchain-ai/langchain/blob/master/libs/langchain/langchain/llms/databricks.py#L86
def get_repl_context() -> Any:
    """Gets the notebook REPL context if running inside a Databricks notebook.
    Returns None otherwise.
    """
    try:
        from dbruntime.databricks_repl_context import get_context

        return get_context()
    except ImportError:
        raise ImportError(
            "Cannot access dbruntime, not running inside a Databricks notebook."
        )


def get_current_username() -> str:
    try:
        from databricks.sdk import WorkspaceClient
        return WorkspaceClient(host=get_repl_context().browserHostName,
                               token=get_repl_context().apiToken).current_user.me().user_name
    except ImportError:
        raise ImportError(
            "Please install databricks-sdk."
        )


def extract_hostname(url):
    parsed_url = urlparse(url)
    return parsed_url.hostname


def ensure_scheme(url):
    if not url.startswith("http") and not url.startswith("https"):
        return f"https://{url}"

class DbTunnel(abc.ABC):

    def __init__(self, port: int, flavor: Flavor):
        self._port = port
        self._flavor = flavor
        import IPython
        self._dbutils = IPython.get_ipython().user_ns["dbutils"]
        self._display_html = IPython.get_ipython().user_ns["displayHTML"]
        self._context = json.loads(self._dbutils.notebook.entry_point.getDbutils().notebook().getContext().toJson())
        self._org_id = self._context["tags"]["orgId"]
        self._cluster_id = self._context["tags"]["clusterId"]
        # need to do this after the context is set
        self._cloud = get_cloud(self._context)
        self._proxy_settings = get_cloud_proxy_settings(self._cloud, self._org_id, self._cluster_id,
                                                        self._port)
        self._loop = None
        self._share = False
        self._share_information = None
        self._share_trigger_callback = None

    @abc.abstractmethod
    def _imports(self):
        pass

    @abc.abstractmethod
    def _run(self):
        pass

    @abc.abstractmethod
    def _display_url(self):
        pass

    @property
    def shared(self):
        return self._share

    def inject_auth(self, host: str = None, token: str = None):
        if os.getenv("DATABRICKS_HOST") is None:
            print("Setting databricks host from context")
            os.environ["DATABRICKS_HOST"] = host or ensure_scheme(get_repl_context().browserHostName)
        if os.getenv("DATABRICKS_TOKEN") is None:
            print("Setting databricks token from context")
            os.environ["DATABRICKS_TOKEN"] = token or get_repl_context().apiToken

        return self

    def inject_sql_warehouse(self, http_path: str, server_hostname: str = None, token: str = None):
        if os.getenv("DATABRICKS_SERVER_HOSTNAME") is None:
            print("Setting databricks server hostname from context")
            os.environ["DATABRICKS_SERVER_HOSTNAME"] = server_hostname or extract_hostname(
                get_repl_context().browserHostName)

        if os.getenv("DATABRICKS_TOKEN") is None:
            print("Setting databricks token from context")
            os.environ["DATABRICKS_TOKEN"] = token or get_repl_context().apiToken

        if os.getenv("DATABRICKS_HTTP_PATH") is None:
            print("Setting databricks warehouse http path")
            os.environ["DATABRICKS_HTTP_PATH"] = http_path

        return self

    def inject_env(self, **kwargs):
        for k, v in kwargs.items():
            if type(v) != str:
                raise ValueError(f"Value for environment variable {k} must be a string")
            print(f"Setting environment variable {k}")
            os.environ[k] = v
        return self

    def run(self):
        self._imports()
        if self._share is True and self._share_trigger_callback is not None:
            import nest_asyncio
            nest_asyncio.apply()
            self._share_trigger_callback()
        if self._share is True and self._share_information is not None:
            print(f"Use this information to publicly access your app: \n{self._share_information.public_url}")
        self._run()

    # right now only ngrok is supported so auth token is required field but in future there may be devtunnels
    def share_to_internet_via_ngrok(self,
                                    *,
                                    ngrok_api_token: str,
                                    ngrok_tunnel_auth_token: str,
                                    kill_existing_processes: bool = True,
                                    kill_all_tunnel_sessions: bool = False,
                                    basic_auth: str = None,  # "databricks:password"
                                    domain: str = None,
                                    oauth_provider: str = None,
                                    oauth_allow_domains: list[str] = None):
        self._share = True
        if kill_existing_processes is True:
            try:
                pkill("ngrok")
            except KeyError:
                print("no running tunnels to kill")
        from dbtunnel.ngrok import NgrokTunnel
        ngrok_tunnel = NgrokTunnel(self._port,
                                   ngrok_tunnel_auth_token,
                                   ngrok_api_token,
                                   basic_auth=basic_auth,
                                   domain=domain,
                                   oauth_provider=oauth_provider,
                                   oauth_allow_domains=oauth_allow_domains,
                                   )

        def ngrok_callback():
            if kill_all_tunnel_sessions is True:
                ngrok_tunnel.kill_existing_sessions()
            ngrok_tunnel.run()

        self._share_trigger_callback = ngrok_callback

        return self

    def display(self):
        pass  # no op because of the annoying flickering
        # self._display_html(self._display_url())



import argparse
import asyncio
from typing import Tuple
from urllib.parse import urlparse

from starlette.types import ASGIApp

from dbtunnel.vendor.asgiproxy.config import BaseURLProxyConfigMixin, ProxyConfig
from dbtunnel.vendor.asgiproxy.context import ProxyContext
from dbtunnel.vendor.asgiproxy.simple_proxy import make_simple_proxy_app

try:
    import uvicorn
except ImportError:
    uvicorn = None


def make_app(upstream_base_url: str) -> Tuple[ASGIApp, ProxyContext]:
    config = type(
        "Config",
        (BaseURLProxyConfigMixin, ProxyConfig),
        {
            "upstream_base_url": upstream_base_url,
            "rewrite_host_header": urlparse(upstream_base_url).netloc,
        },
    )()
    proxy_context = ProxyContext(config)
    app = make_simple_proxy_app(proxy_context)
    return (app, proxy_context)


def main():
    ap = argparse.ArgumentParser()
    ap.add_argument("target")
    ap.add_argument("--port", type=int, default=40404)
    ap.add_argument("--host", type=str, default="0.0.0.0")
    args = ap.parse_args()
    if not uvicorn:
        ap.error(
            "The `uvicorn` ASGI server package is required for the command line client."
        )
    app, proxy_context = make_app(upstream_base_url=args.target)
    try:
        return uvicorn.run(host=args.host, port=int(args.port), app=app)
    finally:
        asyncio.run(proxy_context.close())


if __name__ == "__main__":
    main()



from starlette.types import ASGIApp, Receive, Scope, Send

from dbtunnel.vendor.asgiproxy.context import ProxyContext
from dbtunnel.vendor.asgiproxy.proxies.http import proxy_http
from dbtunnel.vendor.asgiproxy.proxies.websocket import proxy_websocket


def make_simple_proxy_app(
    proxy_context: ProxyContext,
    *,
    proxy_http_handler=proxy_http,
    proxy_websocket_handler=proxy_websocket,
) -> ASGIApp:
    """
    Given a ProxyContext, return a simple ASGI application that can proxy
    HTTP and WebSocket connections.

    The handlers for the protocols can be overridden and/or removed with the
    respective parameters.
    """

    async def app(scope: Scope, receive: Receive, send: Send):  # noqa: ANN201
        if scope["type"] == "lifespan":
            return None  # We explicitly do nothing here for this simple app.

        if scope["type"] == "http" and proxy_http_handler:
            return await proxy_http_handler(
                context=proxy_context, scope=scope, receive=receive, send=send
            )

        if scope["type"] == "websocket" and proxy_websocket_handler:
            return await proxy_websocket_handler(
                context=proxy_context, scope=scope, receive=receive, send=send
            )

        raise NotImplementedError(f"Scope {scope} is not understood")

    return app



import fnmatch
from typing import AsyncGenerator, Union

import aiohttp
from starlette.requests import Request
from starlette.responses import Response, StreamingResponse
from starlette.types import Receive, Scope, Send

from dbtunnel.vendor.asgiproxy.context import ProxyContext
from dbtunnel.vendor.asgiproxy.utils.streams import read_stream_in_chunks

# TODO: make these configurable?
INCOMING_STREAMING_THRESHOLD = 512 * 1024
OUTGOING_STREAMING_THRESHOLD = 1024 * 1024 * 5


def determine_incoming_streaming(request: Request) -> bool:
    if request.method in ("GET", "HEAD"):
        return False

    try:
        return int(request.headers["content-length"]) < INCOMING_STREAMING_THRESHOLD
    except (TypeError, ValueError, KeyError):
        # Malformed or missing content-length header; assume a very large payload
        return True


def determine_outgoing_streaming(proxy_response: aiohttp.ClientResponse) -> bool:
    if proxy_response.status != 200:
        return False
    try:
        return (
            int(proxy_response.headers["content-length"]) > OUTGOING_STREAMING_THRESHOLD
        )
    except (TypeError, ValueError, KeyError):
        # Malformed or missing content-length header; assume a streaming payload
        return True


async def get_proxy_response(
    *,
    context: ProxyContext,
    scope: Scope,
    receive: Receive,
) -> aiohttp.ClientResponse:
    request = Request(scope, receive)
    should_stream_incoming = determine_incoming_streaming(request)
    async with context.semaphore:
        data: Union[None, AsyncGenerator[bytes, None], bytes] = None
        if request.method not in ("GET", "HEAD"):
            if should_stream_incoming:
                data = request.stream()
            else:
                data = await request.body()

        kwargs = context.config.get_upstream_http_options(
            scope=scope, client_request=request, data=data
        )

        return await context.session.request(**kwargs)


async def convert_proxy_response_to_user_response(
    *,
    context: ProxyContext,
    scope: Scope,
    proxy_response: aiohttp.ClientResponse,
) -> Response:
    headers_to_client = context.config.process_upstream_headers(
        scope=scope, proxy_response=proxy_response
    )
    status_to_client = proxy_response.status
    if determine_outgoing_streaming(proxy_response):
        return StreamingResponse(
            content=read_stream_in_chunks(proxy_response.content),
            status_code=status_to_client,
            headers=headers_to_client,  # type: ignore
        )
    response_content = await proxy_response.read()
    new_headers = headers_to_client

    # Forked code
    if response_content is not None and len(response_content) > 0:
        for path_pattern, modify_func in context.config.modify_content.items():
            if fnmatch.fnmatch(scope["path"], path_pattern):
                response_content = modify_func(response_content)
                new_headers = {k:v for k, v in headers_to_client.items()}
                new_headers["Content-Length"] = str(len(response_content))
    return Response(
        content=response_content,
        status_code=status_to_client,
        headers=new_headers,  # type: ignore
    )


async def proxy_http(
    *,
    context: ProxyContext,
    scope: Scope,
    receive: Receive,
    send: Send,
) -> None:

    root_path = scope["root_path"]
    if scope["path"].startswith(root_path):
        scope["path"] = scope["path"].replace(root_path, "")

    proxy_response = await get_proxy_response(
        context=context, scope=scope, receive=receive
    )
    user_response = await convert_proxy_response_to_user_response(
        context=context, scope=scope, proxy_response=proxy_response
    )
    return await user_response(scope, receive, send)



import os
import shutil
import subprocess
import tempfile
from contextlib import contextmanager
from typing import List


@contextmanager
def process_file(input_path):
    temp_dir = tempfile.mkdtemp()
    try:
        # Create a temporary directory

        # Build the destination path in the temporary directory
        temp_file_path = os.path.join(temp_dir, os.path.basename(input_path))

        # Copy the file to the temporary directory
        shutil.copy(input_path, temp_file_path)

        # Yield the temporary file path to the caller
        yield temp_file_path
    finally:
        # Cleanup: Remove the temporary directory and its contents
        shutil.rmtree(temp_dir, ignore_errors=True)


def ensure_python_path(env):
    import sys
    import os
    from pathlib import Path
    for python_version_dir in (Path(sys.executable).parent.parent / "lib").iterdir():
        site_packages = str(python_version_dir / "site-packages")
        py_path = env.get("PYTHONPATH", "")
        if site_packages not in py_path.split(":"):
            env["PYTHONPATH"] = f"{py_path}:{site_packages}"


def execute(cmd: List[str], env, cwd=None, ensure_python_site_packages=True):
    if ensure_python_site_packages:
        ensure_python_path(env)
    import subprocess
    popen = subprocess.Popen(cmd, stdout=subprocess.PIPE, universal_newlines=True, env=env, cwd=cwd)
    for stdout_line in iter(popen.stdout.readline, ""):
        yield stdout_line
    popen.stdout.close()
    return_code = popen.wait()
    if return_code:
        raise subprocess.CalledProcessError(return_code, cmd)


def pkill(process_name):
    try:
        subprocess.run(["pkill", process_name])
        print(f"Processes with name '{process_name}' killed successfully.")
    except subprocess.CalledProcessError as e:
        print(f"Error killing processes: {e}")


def make_asgi_proxy_app(proxy_config):
    from dbtunnel.vendor.asgiproxy.context import ProxyContext
    from dbtunnel.vendor.asgiproxy.simple_proxy import make_simple_proxy_app

    proxy_context = ProxyContext(proxy_config)
    app = make_simple_proxy_app(proxy_context)
    return app



import functools
import re
import threading

from dbtunnel.tunnels import DbTunnel, ProxySettings
from dbtunnel.utils import execute, make_asgi_proxy_app


def make_chainlit_local_proxy_config(url_base_path: str, service_host: str = "0.0.0.0",
                                     service_port: int = 9989):
    from dbtunnel.vendor.asgiproxy.config import BaseURLProxyConfigMixin, ProxyConfig
    proxy_root_path = url_base_path

    def _modify_root(content, root_path):
        list_of_uris = [b"/assets", b"/public", b"/favicon"]
        for uri in list_of_uris:
            content = content.replace(uri, root_path.encode("utf-8") + uri)
        return content

    def _modify_js_content_root_rewrite(content):
        regex_pattern = r'\{path:"\/",element:(\w+)\.jsx\((\w+),\{\}\)\}'

        decoded_content = content.decode("utf-8")
        # Find all matches
        # find the default root function
        matches = re.findall(regex_pattern, decoded_content)
        if matches:
            print(matches)
            jsx_call = matches[0][0]
            func = matches[0][1]  # Assuming there is only one match

            print(f"Found match: {jsx_call}, {func}")
            modified_code = re.sub(r'\{path:"\*",element:.*\.jsx\(.*,\{replace:!0,to:"\/"\}\)\}',
                                   f'{{path:"*",element:{jsx_call}.jsx({func},{{}})}}', decoded_content)
            return modified_code.encode("utf-8")
        else:
            print("No match found.")
            return content

    def _modify_js_bundle(content, root_path):
        list_of_uris = [b"/project/settings", b"/auth/config", b"/ws/socket.io", b"/logo", b"/readme"]
        for uri in list_of_uris:
            content = content.replace(uri, root_path.encode("utf-8") + uri)
        
        content = content.replace(b'to:"/",',f'to:"{root_path}",'.encode("utf-8"))
        content = _modify_js_content_root_rewrite(content)
        return content
    
    def modify_css_bundle(content):
        add_to_end = b" #new-chat-button {display: none;}"
        return content + add_to_end

    def _modify_settings(content, root_path):
        list_of_uris = [b"/public"]
        for uri in list_of_uris:
            content = content.replace(uri, root_path.encode("utf-8") + uri)
        return content

    modify_root = functools.partial(_modify_root, root_path=proxy_root_path)
    modify_js_bundle = functools.partial(_modify_js_bundle, root_path=proxy_root_path)
    modify_settings = functools.partial(_modify_settings, root_path=proxy_root_path)

    config = type(
        "Config",
        (BaseURLProxyConfigMixin, ProxyConfig),
        {
            "upstream_base_url": f"http://{service_host}:{service_port}",
            "rewrite_host_header": f"{service_host}:{service_port}",
            "modify_content": {
                "/": modify_root,
                "": modify_root,
                "*assets/index-*.js": modify_js_bundle,
                "*settings": modify_settings,
                "*assets/index-*.css": modify_css_bundle,
            }
        },
    )()
    return config


class ChainlitAppTunnel(DbTunnel):

    def _imports(self):
        try:
            import chainlit
            import nest_asyncio
        except ImportError as e:
            print("ImportError: Make sure you have chainlit installed. \n"
                  "pip install chainlit nest_asyncio")
            raise e

    def _run(self):
        import os

        chainlit_service_port_no_share = 9090
        if self._share is False:

            url_base_path = self._proxy_settings.url_base_path
            port = self._port

            # nest uvicorn to the ipynotebook asyncio eventloop so restarting kernel kills server
            import nest_asyncio
            nest_asyncio.apply()

            def run_uvicorn_app():
                print("Starting proxy server...")
                app = make_asgi_proxy_app(make_chainlit_local_proxy_config(
                    url_base_path,
                    service_port=chainlit_service_port_no_share
                ))
                import uvicorn
                return uvicorn.run(host="0.0.0.0",
                                   loop="asyncio",
                                   port=int(port),
                                   app=app,
                                   root_path=url_base_path)

            uvicorn_thread = threading.Thread(target=run_uvicorn_app)
            # Start the thread in the background
            uvicorn_thread.start()
            print(f"Use this link to access the UI in Databricks: \n{self._proxy_settings.proxy_url}")

        print("Starting chainlit...", flush=True)

        my_env = os.environ.copy()

        if self._share is False:
            cmd = ["chainlit", "run", self._chainlit_script_path, "-h", "--host", "0.0.0.0", "--port",
                   f"{chainlit_service_port_no_share}"]
        else:
            cmd = ["chainlit", "run", self._chainlit_script_path, "-h", "--host", "0.0.0.0", "--port", f"{self._port}"]

        print(f"Running command: {' '.join(cmd)}")
        for path in execute(cmd, my_env, cwd=self._cwd):
            print(path, end="")

    def __init__(self, chainlit_script_path: str, cwd: str = None, port: int = 8000):
        super().__init__(port, "chainlit")
        self._chainlit_script_path = chainlit_script_path
        self._cwd = cwd

    def _display_url(self):
        return None
